from __future__ import absolute_import

import time

import opentracing as ot
from basictracer import BasicTracer
from basictracer.context import SpanContext

from . import sensor
from . import options as o
from . import recorder as r
from .http_propagator import HTTPPropagator
from .span import InstanaSpan
from .text_propagator import TextPropagator
from .util import generate_id


class InstanaTracer(BasicTracer):
    sensor = None

    def __init__(self, options=o.Options()):
        self.sensor = sensor.get_sensor()
        super(InstanaTracer, self).__init__(
            r.InstanaRecorder(self.sensor), r.InstanaSampler())

        self._propagators[ot.Format.HTTP_HEADERS] = HTTPPropagator()
        self._propagators[ot.Format.TEXT_MAP] = TextPropagator()

    def start_active_span(self,
                          operation_name,
                          child_of=None,
                          references=None,
                          tags=None,
                          start_time=None,
                          ignore_active_span=False,
                          finish_on_close=True):

        # create a new Span
        span = self.start_span(
            operation_name=operation_name,
            child_of=child_of,
            references=references,
            tags=tags,
            start_time=start_time,
            ignore_active_span=ignore_active_span,
        )

        return self.scope_manager.activate(span, finish_on_close)

    def start_span(self,
                   operation_name=None,
                   child_of=None,
                   references=None,
                   tags=None,
                   start_time=None,
                   ignore_active_span=False):
        "Taken from BasicTracer so we can override generate_id calls to ours"

        start_time = time.time() if start_time is None else start_time

        # See if we have a parent_ctx in `references`
        parent_ctx = None
        if child_of is not None:
            parent_ctx = (
                child_of if isinstance(child_of, ot.SpanContext)
                else child_of.context)
        elif references is not None and len(references) > 0:
            # TODO only the first reference is currently used
            parent_ctx = references[0].referenced_context

        # retrieve the active SpanContext
        if not ignore_active_span and parent_ctx is None:
            scope = self.scope_manager.active
            if scope is not None:
                parent_ctx = scope.span.context

        # Assemble the child ctx
        gid = generate_id()
        ctx = SpanContext(span_id=gid)
        if parent_ctx is not None:
            if parent_ctx._baggage is not None:
                ctx._baggage = parent_ctx._baggage.copy()
            ctx.trace_id = parent_ctx.trace_id
            ctx.sampled = parent_ctx.sampled
        else:
            ctx.trace_id = gid
            ctx.sampled = self.sampler.sampled(ctx.trace_id)

        # Tie it all together
        return InstanaSpan(
            self,
            operation_name=operation_name,
            context=ctx,
            parent_id=(None if parent_ctx is None else parent_ctx.span_id),
            tags=tags,
            start_time=start_time)

    def inject(self, span_context, format, carrier):
        if format in self._propagators:
            self._propagators[format].inject(span_context, carrier)
        else:
            raise ot.UnsupportedFormatException()

    def extract(self, format, carrier):
        if format in self._propagators:
            return self._propagators[format].extract(carrier)
        else:
            raise ot.UnsupportedFormatException()

    def handle_fork(self):
        self.sensor.handle_fork()
        self.recorder = r.InstanaRecorder(self.sensor)


def init(options):
    """
    Deprecated.
        No longer in use.
        To be removed in next major release.
    """
    return internal_tracer


def get_tracer():
    return internal_tracer


# The global OpenTracing compatible tracer used internally by
# this package.
#
# Usage example:
#
# import instana
# instana.internal_tracer.start_span(...)
#
internal_tracer = InstanaTracer()

# Set ourselves as the tracer.
ot.tracer = internal_tracer



import logging
import os


class Options(object):
    service = ''
    agent_host = ''
    agent_port = 0
    log_level = logging.WARN

    def __init__(self, **kwds):
        """ Initialize Options
        Respect any environment variables that may be set.
        """
        if "INSTANA_DEV" in os.environ:
            self.log_level = logging.DEBUG

        if "INSTANA_AGENT_IP" in os.environ:
            # Deprecated: INSTANA_AGENT_IP environment variable
            # To be removed in a future version
            self.agent_host = os.environ["INSTANA_AGENT_IP"]

        if "INSTANA_AGENT_HOST" in os.environ:
            self.agent_host = os.environ["INSTANA_AGENT_HOST"]

        if "INSTANA_AGENT_PORT" in os.environ:
            self.agent_port = os.environ["INSTANA_AGENT_PORT"]

        self.__dict__.update(kwds)



import copy
import gc as gc_
import os
import platform
import resource
import sys
import threading
import time
from types import ModuleType

from pkg_resources import get_distribution, DistributionNotFound

from .agent_const import AGENT_DATA_URL
from .log import logger as log


class Snapshot(object):
    name = None
    version = None
    f = None  # flavor: CPython, Jython, IronPython, PyPy
    a = None  # architecture: i386, x86, x86_64, AMD64
    versions = None
    djmw = []

    def __init__(self, **kwds):
        self.__dict__.update(kwds)

    def to_dict(self):
        kvs = dict()
        kvs['name'] = self.name
        kvs['version'] = self.version
        kvs['f'] = self.f  # flavor
        kvs['a'] = self.a  # architecture
        kvs['versions'] = self.versions
        kvs['djmw'] = list(self.djmw)
        return kvs


class GC(object):
    collect0 = 0
    collect1 = 0
    collect2 = 0
    threshold0 = 0
    threshold1 = 0
    threshold2 = 0

    def __init__(self, **kwds):
        self.__dict__.update(kwds)

    def to_dict(self):
        return self.__dict__


class Metrics(object):
    ru_utime = .0
    ru_stime = .0
    ru_maxrss = 0
    ru_ixrss = 0
    ru_idrss = 0
    ru_isrss = 0
    ru_minflt = 0
    ru_majflt = 0
    ru_nswap = 0
    ru_inblock = 0
    ru_oublock = 0
    ru_msgsnd = 0
    ru_msgrcv = 0
    ru_nsignals = 0
    ru_nvcs = 0
    ru_nivcsw = 0
    dummy_threads = 0
    alive_threads = 0
    daemon_threads = 0
    gc = None

    def __init__(self, **kwds):
        self.__dict__.update(kwds)

    def delta_data(self, delta):
        data = self.__dict__
        if delta is None:
            return data

        unchanged_items = set(data.items()) & set(delta.items())
        for x in unchanged_items:
            data.pop(x[0])

        return data

    def to_dict(self):
        return self.__dict__


class EntityData(object):
    pid = 0
    snapshot = None
    metrics = None

    def __init__(self, **kwds):
        self.__dict__.update(kwds)

    def to_dict(self):
        return self.__dict__


class Meter(object):
    SNAPSHOT_PERIOD = 600
    snapshot_countdown = 35
    sensor = None
    last_usage = None
    last_collect = None
    timer = None
    last_metrics = None
    djmw = None

    def __init__(self, sensor):
        self.sensor = sensor

    def run(self):
        self.timer = threading.Thread(target=self.collect_and_report)
        self.timer.daemon = True
        self.timer.name = "Instana Metric Collection"
        self.timer.start()

    def collect_and_report(self):
        while 1:
            self.process()
            if (self.sensor.agent.is_timed_out()):
                log.warn("Host agent offline for >1 min.  Going to sit in a corner...")
                self.sensor.agent.reset()
                break
            time.sleep(1)

    def process(self):
        if self.sensor.agent.can_send():
            self.snapshot_countdown = self.snapshot_countdown - 1
            ss = None
            cm = self.collect_metrics()

            if self.snapshot_countdown < 1:
                self.snapshot_countdown = self.SNAPSHOT_PERIOD
                ss = self.collect_snapshot()
                md = copy.deepcopy(cm).delta_data(None)
            else:
                md = copy.deepcopy(cm).delta_data(self.last_metrics)

            ed = EntityData(pid=self.sensor.agent.from_.pid, snapshot=ss, metrics=md)
            url = self.sensor.agent.make_url(AGENT_DATA_URL)
            self.sensor.agent.request(url, "POST", ed)
            self.last_metrics = cm.__dict__

    def collect_snapshot(self):
        try:
            if "FLASK_APP" in os.environ:
                appname = os.environ["FLASK_APP"]
            elif "DJANGO_SETTINGS_MODULE" in os.environ:
                appname = os.environ["DJANGO_SETTINGS_MODULE"].split('.')[0]
            elif os.path.basename(sys.argv[0]) == '' and sys.stdout.isatty():
                appname = "Interactive Console"
            else:
                if os.path.basename(sys.argv[0]) == '':
                    appname = os.path.basename(sys.executable)
                else:
                    appname = os.path.basename(sys.argv[0])

            s = Snapshot(name=appname, version=platform.version(),
                         f=platform.python_implementation(),
                         a=platform.architecture()[0],
                         djmw=self.djmw)
            s.version = sys.version
            s.versions = self.collect_modules()
        except Exception as e:
            log.debug(e.message)
        else:
            return s

    def jsonable(self, value):
        try:
            if callable(value):
                result = value()
            elif type(value) is ModuleType:
                result = value
            else:
                result = value
            return str(result)
        except Exception as e:
            log.debug(e)

    def collect_modules(self):
        try:
            res = {}
            m = sys.modules
            for k in m:
                # Don't report submodules (e.g. django.x, django.y, django.z)
                # Skip modules that begin with underscore
                if ('.' in k) or k[0] == '_':
                    continue
                if m[k]:
                    try:
                        d = m[k].__dict__
                        if "version" in d and d["version"]:
                            res[k] = self.jsonable(d["version"])
                        elif "__version__" in d and d["__version__"]:
                            res[k] = self.jsonable(d["__version__"])
                        else:
                            res[k] = get_distribution(k).version
                    except DistributionNotFound:
                        pass
                    except Exception:
                        log.debug("collect_modules: could not process module: %s" % k)

        except Exception:
            log.debug("collect_modules", exc_info=True)
        else:
            return res

    def collect_metrics(self):
        u = resource.getrusage(resource.RUSAGE_SELF)
        if gc_.isenabled():
            c = list(gc_.get_count())
            th = list(gc_.get_threshold())
            g = GC(collect0=c[0] if not self.last_collect else c[0] - self.last_collect[0],
                   collect1=c[1] if not self.last_collect else c[
                       1] - self.last_collect[1],
                   collect2=c[2] if not self.last_collect else c[
                       2] - self.last_collect[2],
                   threshold0=th[0],
                   threshold1=th[1],
                   threshold2=th[2])

        thr = threading.enumerate()
        daemon_threads = [tr.daemon is True for tr in thr].count(True)
        alive_threads = [tr.daemon is False for tr in thr].count(True)
        dummy_threads = [type(tr) is threading._DummyThread for tr in thr].count(True)

        m = Metrics(ru_utime=u[0] if not self.last_usage else u[0] - self.last_usage[0],
                    ru_stime=u[1] if not self.last_usage else u[1] - self.last_usage[1],
                    ru_maxrss=u[2],
                    ru_ixrss=u[3],
                    ru_idrss=u[4],
                    ru_isrss=u[5],
                    ru_minflt=u[6] if not self.last_usage else u[6] - self.last_usage[6],
                    ru_majflt=u[7] if not self.last_usage else u[7] - self.last_usage[7],
                    ru_nswap=u[8] if not self.last_usage else u[8] - self.last_usage[8],
                    ru_inblock=u[9] if not self.last_usage else u[9] - self.last_usage[9],
                    ru_oublock=u[10] if not self.last_usage else u[10] - self.last_usage[10],
                    ru_msgsnd=u[11] if not self.last_usage else u[11] - self.last_usage[11],
                    ru_msgrcv=u[12] if not self.last_usage else u[12] - self.last_usage[12],
                    ru_nsignals=u[13] if not self.last_usage else u[13] - self.last_usage[13],
                    ru_nvcs=u[14] if not self.last_usage else u[14] - self.last_usage[14],
                    ru_nivcsw=u[15] if not self.last_usage else u[15] - self.last_usage[15],
                    alive_threads=alive_threads,
                    dummy_threads=dummy_threads,
                    daemon_threads=daemon_threads,
                    gc=g)

        self.last_usage = u
        if gc_.isenabled():
            self.last_collect = c

        return m



from basictracer.context import SpanContext
from basictracer.span import BasicSpan


class InstanaSpan(BasicSpan):
    def finish(self, finish_time=None):
        super(InstanaSpan, self).finish(finish_time)

    def log_exception(self, e):
        if hasattr(e, 'message') and len(e.message):
            self.log_kv({'message': e.message})
        elif hasattr(e, '__str__'):
            self.log_kv({'message': e.__str__()})
        else:
            self.log_kv({'message': str(e)})

        self.set_tag("error", True)
        ec = self.tags.get('ec', 0)
        self.set_tag("ec", ec+1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import opentracing.ext.tags as ext
from flask import Flask, redirect
from instana.wsgi import iWSGIMiddleware
from wsgiref.simple_server import make_server
from instana.tracer import internal_tracer as tracer


app = Flask(__name__)
app.debug = False
app.use_reloader = False

wsgi_app = iWSGIMiddleware(app.wsgi_app)
flask_server = make_server('127.0.0.1', 5000, wsgi_app)


@app.route("/")
def hello():
    return "<center><h1>🐍 Hello Stan! 🦄</h1></center>"


@app.route("/complex")
def gen_opentracing():
    with tracer.start_active_span('asteroid') as pscope:
        pscope.span.set_tag(ext.COMPONENT, "Python simple example app")
        pscope.span.set_tag(ext.SPAN_KIND, ext.SPAN_KIND_RPC_SERVER)
        pscope.span.set_tag(ext.PEER_HOSTNAME, "localhost")
        pscope.span.set_tag(ext.HTTP_URL, "/python/simple/one")
        pscope.span.set_tag(ext.HTTP_METHOD, "GET")
        pscope.span.set_tag(ext.HTTP_STATUS_CODE, 200)
        pscope.span.log_kv({"foo": "bar"})

        with tracer.start_active_span('spacedust', child_of=pscope.span) as cscope:
            cscope.span.set_tag(ext.SPAN_KIND, ext.SPAN_KIND_RPC_CLIENT)
            cscope.span.set_tag(ext.PEER_HOSTNAME, "localhost")
            cscope.span.set_tag(ext.HTTP_URL, "/python/simple/two")
            cscope.span.set_tag(ext.HTTP_METHOD, "POST")
            cscope.span.set_tag(ext.HTTP_STATUS_CODE, 204)
            cscope.span.set_baggage_item("someBaggage", "someValue")

    return "<center><h1>🐍 Generated some OT spans... 🦄</h1></center>"


@app.route("/301")
def threehundredone():
    return redirect('/', code=301)


@app.route("/302")
def threehundredtwo():
    return redirect('/', code=302)


@app.route("/400")
def fourhundred():
    return "Simulated Bad Request", 400


@app.route("/405")
def fourhundredfive():
    return "Simulated Method not allowed", 405


@app.route("/500")
def fivehundred():
    return "Simulated Internal Server Error", 500


@app.route("/504")
def fivehundredfour():
    return "Simulated Gateway Timeout", 504


@app.route("/exception")
def exception():
    raise Exception('fake error')


if __name__ == '__main__':
    flask_server.serve_forever()



# This is a wrapper for PEP-0249: Python Database API Specification v2.0
import opentracing.ext.tags as ext
import wrapt

from ..tracer import internal_tracer
from ..log import logger


class CursorWrapper(wrapt.ObjectProxy):
    __slots__ = ('_module_name', '_connect_params', '_cursor_params')

    def __init__(self, cursor, module_name,
                 connect_params=None, cursor_params=None):
        super(CursorWrapper, self).__init__(wrapped=cursor)
        self._module_name = module_name
        self._connect_params = connect_params
        self._cursor_params = cursor_params

    def _collect_kvs(self, span, sql):
        try:
            span.set_tag(ext.SPAN_KIND, 'exit')
            span.set_tag(ext.DATABASE_INSTANCE, self._connect_params[1]['db'])
            span.set_tag(ext.DATABASE_STATEMENT, sql)
            span.set_tag(ext.DATABASE_TYPE, 'mysql')
            span.set_tag(ext.DATABASE_USER, self._connect_params[1]['user'])
            span.set_tag('host', "%s:%s" %
                         (self._connect_params[1]['host'],
                          self._connect_params[1]['port']))
        except Exception as e:
            logger.debug(e)
        finally:
            return span

    def execute(self, sql, params=None):
        parent_span = internal_tracer.active_span

        # If we're not tracing, just return
        if parent_span is None:
            return self.__wrapped__.execute(sql, params)

        with internal_tracer.start_active_span(self._module_name, child_of=parent_span) as scope:
            try:
                self._collect_kvs(scope.span, sql)

                result = self.__wrapped__.execute(sql, params)
            except Exception as e:
                if scope.span:
                    scope.span.log_exception(e)
                raise
            else:
                return result

    def executemany(self, sql, seq_of_parameters):
        parent_span = internal_tracer.active_span

        # If we're not tracing, just return
        if parent_span is None:
            return self.__wrapped__.executemany(sql, seq_of_parameters)

        with internal_tracer.start_active_span(self._module_name, child_of=parent_span) as scope:
            try:
                self._collect_kvs(scope.span, sql)

                result = self.__wrapped__.executemany(sql, seq_of_parameters)
            except Exception as e:
                if scope.span:
                    scope.span.log_exception(e)
                raise
            else:
                return result

    def callproc(self, proc_name, params):
        parent_span = internal_tracer.active_span

        # If we're not tracing, just return
        if parent_span is None:
            return self.__wrapped__.execute(proc_name, params)

        with internal_tracer.start_active_span(self._module_name, child_of=parent_span) as scope:
            try:
                self._collect_kvs(scope.span, proc_name)

                result = self.__wrapped__.callproc(proc_name, params)
            except Exception as e:
                if scope.span:
                    scope.span.log_exception(e)
                raise
            else:
                return result


class ConnectionWrapper(wrapt.ObjectProxy):
    __slots__ = ('_module_name', '_connect_params')

    def __init__(self, connection, module_name, connect_params):
        super(ConnectionWrapper, self).__init__(wrapped=connection)
        self._module_name = module_name
        self._connect_params = connect_params

    def cursor(self, *args, **kwargs):
        return CursorWrapper(
            cursor=self.__wrapped__.cursor(*args, **kwargs),
            module_name=self._module_name,
            connect_params=self._connect_params,
            cursor_params=(args, kwargs) if args or kwargs else None)

    def begin(self):
        return self.__wrapped__.begin()

    def commit(self):
        return self.__wrapped__.commit()

    def rollback(self):
        return self.__wrapped__.rollback()


class ConnectionFactory(object):
    def __init__(self, connect_func, module_name):
        self._connect_func = connect_func
        self._module_name = module_name
        self._wrapper_ctor = ConnectionWrapper

    def __call__(self, *args, **kwargs):
        connect_params = (args, kwargs) if args or kwargs else None

        return self._wrapper_ctor(
            connection=self._connect_func(*args, **kwargs),
            module_name=self._module_name,
            connect_params=connect_params)



# vim: set fileencoding=UTF-8 :
import logging
from wsgiref.simple_server import make_server

from spyne import (Application, Fault, Integer, Iterable, ServiceBase, Unicode,
                   rpc)
from spyne.protocol.soap import Soap11
from spyne.server.wsgi import WsgiApplication

from instana.wsgi import iWSGIMiddleware

# Simple in test suite SOAP server to test suds client instrumentation against.
# Configured to listen on localhost port 4132
# WSDL: http://localhost:4232/?wsdl


class StanSoapService(ServiceBase):
    @rpc(Unicode, Integer, _returns=Iterable(Unicode))
    def ask_question(ctx, question, answer):
        """Ask Stan a question!
        <b>Ask Stan questions as a Service</b>

        @param name the name to say hello to
        @param times the number of times to say hello
        @return the completed array
        """

        yield u'To an artificial mind, all reality is virtual. How do they know that the real world isn\'t just another simulation? How do you?'

    @rpc()
    def server_exception(ctx):
        raise Exception("Server side exception example.")

    @rpc()
    def server_fault(ctx):
        raise Fault("Server", "Server side fault example.")

    @rpc()
    def client_fault(ctx):
        raise Fault("Client", "Client side fault example")


# logging.basicConfig(level=logging.WARN)
logging.getLogger('suds').setLevel(logging.WARN)
logging.getLogger('suds.resolver').setLevel(logging.WARN)
logging.getLogger('spyne.protocol.xml').setLevel(logging.WARN)
logging.getLogger('spyne.model.complex').setLevel(logging.WARN)
logging.getLogger('spyne.interface._base').setLevel(logging.WARN)
logging.getLogger('spyne.interface.xml').setLevel(logging.WARN)
logging.getLogger('spyne.util.appreg').setLevel(logging.WARN)

app = Application([StanSoapService], 'instana.tests.app.ask_question',
                  in_protocol=Soap11(validator='lxml'), out_protocol=Soap11())

# Use Instana middleware so we can test context passing and Soap server traces.
wsgi_app = iWSGIMiddleware(WsgiApplication(app))
soapserver = make_server('127.0.0.1', 4132, wsgi_app)

if __name__ == '__main__':
    soapserver.serve_forever()



from __future__ import absolute_import

from . import log
from .agent import Agent
from .meter import Meter
from .options import Options


class Sensor(object):
    options = None
    meter = None
    agent = None

    def __init__(self, options):
        self.set_options(options)
        log.init(options.log_level)
        self.agent = Agent(self)
        self.meter = Meter(self)

        log.debug("initialized sensor")

    def set_options(self, options):
        self.options = options
        if not self.options:
            self.options = Options()

    def handle_fork(self):
        self.agent = Agent(self)
        self.meter = Meter(self)

def get_sensor():
    return global_sensor

# For any given Python process, we only want one sensor as multiple would
# collect/report metrics in duplicate, triplicate etc..
#
# Usage example:
#
# import instana
# instana.global_sensor
#
global_sensor = Sensor(Options())



from __future__ import absolute_import

from distutils.version import LooseVersion

import opentracing
import opentracing.ext.tags as ext
import wrapt

from ..log import logger
from ..tracer import internal_tracer

try:
    import suds # noqa

    if (LooseVersion(suds.version.__version__) <= LooseVersion('0.6')):
        class_method = 'SoapClient.send'
    else:
        class_method = '_SoapClient.send'

    @wrapt.patch_function_wrapper('suds.client', class_method)
    def send_with_instana(wrapped, instance, args, kwargs):
        parent_span = internal_tracer.active_span

        # If we're not tracing, just return
        if parent_span is None:
            return wrapped(*args, **kwargs)

        with internal_tracer.start_active_span("soap", child_of=parent_span) as scope:
            try:
                scope.span.set_tag('soap.action', instance.method.name)
                scope.span.set_tag(ext.HTTP_URL, instance.method.location)
                scope.span.set_tag(ext.HTTP_METHOD, 'POST')

                internal_tracer.inject(scope.span.context, opentracing.Format.HTTP_HEADERS, instance.options.headers)

                rv = wrapped(*args, **kwargs)

            except Exception as e:
                scope.span.log_exception(e)
                scope.span.set_tag(ext.HTTP_STATUS_CODE, 500)
                raise
            else:
                scope.span.set_tag(ext.HTTP_STATUS_CODE, 200)
                return rv

    logger.debug("Instrumenting suds-jurko")
except ImportError:
    pass



import opentracing as ot

from instana import options, tracer

# This file is the hook for autoinstrumenation.
# Here, we should:
#  1. Make sure instana sensor is not already active in the process
#  2. Activate properly
#    a. Runtime metrics
#    b. Detect and instrument framework
#    c. Detect and instrument any libraries

opts = options.Options()
ot.tracer = tracer.InstanaTracer(opts)



import logging as log
import os

logger = log.getLogger('instana')

def init(level):
    ch = log.StreamHandler()
    f = log.Formatter('%(asctime)s: %(process)d %(levelname)s %(name)s: %(message)s')
    ch.setFormatter(f)
    logger.addHandler(ch)
    if "INSTANA_DEV" in os.environ:
        logger.setLevel(log.DEBUG)
    else:
        logger.setLevel(level)


def debug(s, *args):
    logger.debug("%s %s", s, ' '.join(args))


def info(s, *args):
    logger.info("%s %s", s, ' '.join(args))


def warn(s, *args):
    logger.warn("%s %s", s, ' '.join(args))


def error(s, *args):
    logger.error("%s %s", s, ' '.join(args))



"""
This module provides a client for the Instana REST API.

Use of this client requires the URL of your Instana account dashboard
and an API token.  The API token can be generated in your dashboard under
Settings > Access Control > API Tokens.

See the associated REST API documentation here:
https://documenter.getpostman.com/view/1527374/instana-api/2TqWQh#intro

The API currently uses the requests package to make the REST calls to the API.
As such, requests response objects are returned from API calls.
"""
import os
import sys
import json
import time
import certifi
import urllib3
from .log import logger as log


PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

if PY2:
    from urllib import urlencode
    import urllib3.contrib.pyopenssl
    urllib3.contrib.pyopenssl.inject_into_urllib3()
else:
    import urllib3
    from urllib.parse import urlencode


# For use with the Token related API calls
token_config = {
    "id": "",
    "name": "",
    "canConfigureServiceMapping": False,
    "canConfigureEumApplications": False,
    "canConfigureUsers": False,
    "canInstallNewAgents": False,
    "canSeeUsageInformation": False,
    "canConfigureIntegrations": False,
    "canSeeOnPremLicenseInformation": False,
    "canConfigureRoles": False,
    "canConfigureCustomAlerts": False,
    "canConfigureApiTokens": False,
    "canConfigureAgentRunMode": False,
    "canViewAuditLog": False,
    "canConfigureObjectives": False
}

# For use with the Bindings related API calls
binding_config = {
  "id": "1",
  "enabled": True,
  "triggering": False,
  "severity": 5,
  "text": "text",
  "description": "desc",
  "expirationTime": 60000,
  "query": "",
  "ruleIds": [
    "2"
  ]
}

# For use with the Rule related API calls
rule_config = {
  "id": "1",
  "name": "test rule",
  "entityType": "mariaDbDatabase",
  "metricName": "status.MAX_USED_CONNECTIONS",
  "rollup": 1000,
  "window": 60000,
  "aggregation": "avg",
  "conditionOperator": ">=",
  "conditionValue": 10
}

# For use with the Role related API calls
role_config = {
    "id": "1",
    "name": "Developer",
    "implicitViewFilter": "",
    "canConfigureServiceMapping": True,
    "canConfigureEumApplications": True,
    "canConfigureUsers": False,
    "canInstallNewAgents": False,
    "canSeeUsageInformation": False,
    "canConfigureIntegrations": False,
    "canSeeOnPremLicenseInformation": False,
    "canConfigureRoles": False,
    "canConfigureCustomAlerts": False,
    "canConfigureApiTokens": False,
    "canConfigureAgentRunMode": False,
    "canViewAuditLog": False,
    "canConfigureObjectives": False
}


class APIClient(object):
    """
    The Python client to the Instana REST API.

    This client supports the use of environment variables.  These environment variables
    will override any passed in options:

        INSTANA_API_TOKEN=asdffdsa
        INSTANA_BASE_URL=https://test-test.instana.io

    Example usage:
        from instana.api import APIClient
        c = APIClient(base_url="https://test-test.instana.io", api_token='asdffdsa')

        # Retrieve the current application view
        x = c.application_view()
        x.json()

        # Retrieve snapshots results from a query
        y = c.snapshots("entity.selfType:webService entity.service.name:\"pwpush.com\"")
    """
    base_url = None
    api_token = None

    def __init__(self, **kwds):
        for key in kwds:
            self.__dict__[key] = kwds[key]

        if "INSTANA_API_TOKEN" in os.environ:
            self.api_token = os.environ["INSTANA_API_TOKEN"]

        if "INSTANA_BASE_URL" in os.environ:
            self.base_url = os.environ["INSTANA_BASE_URL"]

        if self.base_url is None or self.api_token is None:
            log.warn("APIClient: API token or Base URL not set.  No-op mode")
        else:
            self.api_key = "apiToken %s" % self.api_token
            self.headers = {'Authorization': self.api_key}
            self.http = urllib3.PoolManager(cert_reqs='CERT_REQUIRED',
                                            ca_certs=certifi.where())

    def ts_now(self):
        return int(round(time.time() * 1000))

    def build_url(self, path, query_args):
        if self.base_url and self.api_token:
            url = self.base_url + path
        else:
            url = ""

        if query_args:
            encoded_args = urlencode(query_args)
            url = url + '?' + encoded_args
        return url

    def get(self, path, query_args=None):
        if self.base_url and self.api_token:
            url = self.build_url(path, query_args)
            return self.http.request('GET', url, headers=self.headers)

    def put(self, path, query_args=None, payload=''):
        if self.base_url and self.api_token:
            url = self.build_url(path, query_args)
            encoded_data = json.dumps(payload).encode('utf-8')
            post_headers = self.headers
            post_headers['Content-Type'] = 'application/json'
            return self.http.request('PUT', url, body=encoded_data, headers=post_headers)

    def post(self, path, query_args=None, payload=''):
        if self.base_url and self.api_token:
            url = self.build_url(path, query_args)
            encoded_data = json.dumps(payload).encode('utf-8')
            post_headers = self.headers
            post_headers['Content-Type'] = 'application/json'
            return self.http.request('POST', url, body=encoded_data, headers=post_headers)

    def delete(self, path, query_args):
        if self.base_url and self.api_token:
            url = self.build_url(path, query_args)
            return self.http.request('DELETE', url, headers=self.headers)

    def tokens(self):
        return self.get('/api/apiTokens')

    def token(self, token):
        return self.get('/api/apiTokens/%s' % token)

    def delete_token(self, token):
        return self.delete('/api/apiTokens/%s' % token)

    def upsert_token(self, token_config):
        return self.put('/api/apiTokens/%s' % token_config["id"], payload=token_config)

    def audit_log(self):
        return self.get('/api/auditlog')

    def eum_apps(self):
        return self.get('/api/eumApps')

    def create_eum_app(self, name):
        return self.post('/api/eumApps', payload={'name': name})

    def rename_eum_app(self, eum_app_id, new_name):
        return self.put('/api/eumApps/%s' % (eum_app_id), payload={'name': new_name})

    def delete_eum_app(self, eum_app_id):
        return self.delete('/api/eumApps/%s' % eum_app_id)

    def events(self, window_size=300000, to=None):
        if to is None:
            to = self.ts_now()
        return self.get('/api/events/', query_args={'windowsize': window_size, 'to': to})

    def event(self, event_id):
        return self.get('/api/events/%s' % event_id)

    def metrics(self, metric_name, ts_from, ts_to, aggregation, snapshot_id, rollup):
        params = {'metric': metric_name,
                  'from': ts_from,
                  'to': ts_to,
                  'aggregation': aggregation,
                  'snapshotId': snapshot_id,
                  'rollup': rollup}
        return self.get('/api/metrics', query_args=params)

    def metric(self, metric_name, timestamp, aggregation, snapshot_id, rollup):
        params = {'metric': metric_name,
                  'time': timestamp,
                  'aggregation': aggregation,
                  'snapshotId': snapshot_id,
                  'rollup': rollup}
        return self.get('/api/metric', query_args=params)

    def rule_bindings(self):
        return self.get('/api/ruleBindings')

    def rule_binding(self, rule_binding_id):
        return self.get('/api/ruleBindings/%s' % rule_binding_id)

    def upsert_rule_binding(self, rule_binding_config):
        path = '/api/ruleBindings/%s' % rule_binding_config["id"]
        return self.put(path, rule_binding_config)

    def delete_rule_binding(self, rule_binding_id):
        return self.detel('/api/ruleBindings/%s' % rule_binding_id)

    def rules(self):
        return self.get('/api/rules')

    def rule(self, rule_id):
        return self.get('/api/rules/%s' % rule_id)

    def upsert_rule(self, rule_config):
        path = '/api/rules/%s' % rule_config["id"]
        return self.put(path, rule_config)

    def delete_rule(self, rule_id):
        return self.delete('/api/rules/%s' % rule_id)

    def search_fields(self):
        return self.get('/api/searchFields')

    def service_extraction_configs(self):
        return self.get('/api/serviceExtractionConfigs')

    def upsert_service_extraction_configs(self, service_extraction_config):
        path = '/api/serviceExtractionConfigs/%s' % service_extraction_config["id"]
        return self.put(path, service_extraction_config)

    def snapshot(self, id, timestamp=None):
        if timestamp is None:
            timestamp = self.ts_now()

        params = {'time': timestamp}
        path = "/api/snapshots/%s" % id
        return self.get(path, query_args=params)

    def snapshots(self, query, timestamp=None, size=5):
        if timestamp is None:
            timestamp = self.ts_now()

        params = {'time': timestamp, 'q': query, 'size': size}
        path = "/api/snapshots"
        return self.get(path, query_args=params)

    def trace(self, trace_id):
        return self.get('/api/traces/%d' % trace_id)

    def traces_by_timeframe(self, query, window_size, ts_to, sort_by='ts', sort_mode='asc'):
        params = {'windowsize': window_size,
                  'to': ts_to,
                  'sortBy': sort_by,
                  'sortMode': sort_mode,
                  'query': query}
        return self.get('/api/traces', query_args=params)

    def roles(self):
        return self.get('/api/roles')

    def role(self, role_id):
        return self.get('/api/roles/%s' % role_id)

    def upsert_role(self, role_config):
        path = '/api/roles/%s' % role_config["id"]
        return self.put(path, payload=role_config)

    def delete_role(self, role_id):
        return self.delete('/api/roles/%s' % role_id)

    def users(self):
        return self.get('/api/tenant/users/overview')

    def set_user_role(self, user_id, role_id):
        return self.put('/api/tenant/users/%s/role' % user_id,
                        query_args={'roleId': role_id})

    def remove_user_from_tenant(self, user_id):
        return self.delete('/api/tenant/users/%s' % user_id)

    def invite_user(self, email, role_id):
        return self.post('/api/tenant/users/invitations',
                         query_args={'email', email, 'roleId', role_id})

    def revoke_pending_invitation(self, email):
        return self.delete('/api/tenant/users/invitations',
                           query_args={'email': email})

    def application_view(self):
        return self.get('/api/graph/views/application')

    def infrastructure_view(self):
        return self.get('/api/graph/views/infrastructure')

    def usage(self):
        return self.get('/api/usage/')

    def usage_for_month(self, year, month):
        return self.get('/api/usage/%d/%d' % (month, year))

    def usage_for_day(self, year, month, day):
        return self.get('/api/usage/%d/%d/%d' % (day, month, year))

    def average_number_of_hosts_for_month(self, year, month):
        return self.get('/api/usage/hosts/%d/%d' % (month, year))

    def average_number_of_hosts_for_day(self, year, month, day):
        return self.get('/api/usage/hosts/%d/%d/%d' % (month, year, day))



# coding: utf-8
from distutils.version import LooseVersion

from setuptools import find_packages, setup


def check_setuptools():
    import pkg_resources
    st_version = pkg_resources.get_distribution('setuptools').version
    if LooseVersion(st_version) < LooseVersion('20.2.2'):
        exit('The Instana sensor requires a newer verion of `setuptools` (>=20.2.2).\n'
             'Please run `pip install --upgrade setuptools` to upgrade. \n'
             '  and then try the install again.\n'
             'Also:\n'
             '  `pip show setuptools` - shows the current version\n'
             '  To see the setuptool releases: \n'
             '    https://setuptools.readthedocs.io/en/latest/history.html')


check_setuptools()

setup(name='instana',
      version='1.1.3',
      download_url='https://github.com/instana/python-sensor',
      url='https://www.instana.com/',
      license='MIT',
      author='Instana Inc.',
      author_email='peter.lombardo@instana.com',
      description='🐍 Python Distributed Tracing & Metrics Sensor for Instana',
      packages=find_packages(exclude=['tests', 'examples']),
      long_description="The instana package collects and reports Python metrics and distibuted \
traces to your Instana dashboard.",
      zip_safe=False,
      install_requires=['autowrapt>=1.0',
                        'basictracer>=3.0.0',
                        'certifi>=2018.4.16',
                        'fysom>=2.1.2',
                        'opentracing>=2.0.0',
                        'urllib3>=1.18.1'],
      entry_points={
                    'instana':  ['string = instana:load'],
                    'flask':    ['flask = instana.flaskana:hook'],
                    'runtime':  ['string = instana:load'],  # deprecated: use same as 'instana'
                    'django':   ['string = instana:load'],  # deprecated: use same as 'instana'
                    'django19': ['string = instana:load'],  # deprecated: use same as 'instana'
                    },
      extras_require={
        'test': [
            'django>=1.11',
            'nose>=1.0',
            'flask>=0.12.2',
            'lxml>=3.4',
            'MySQL-python>=1.2.5;python_version<="2.7"',
            'pyOpenSSL>=16.1.0;python_version<="2.7"',
            'requests>=2.17.1',
            'urllib3[secure]>=1.15',
            'spyne>=2.9',
            'suds-jurko>=0.6'
        ],
      },
      test_suite='nose.collector',
      keywords=['performance', 'opentracing', 'metrics', 'monitoring', 'tracing', 'distributed-tracing'],
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Django',
        'Framework :: Flask',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware',
        'Topic :: System :: Monitoring',
        'Topic :: System :: Networking :: Monitoring',
        'Topic :: Software Development :: Libraries :: Python Modules'])



from __future__ import absolute_import

import os
from pkg_resources import get_distribution

"""
The Instana package has two core components: the sensor and the tracer.

The sensor is individual to each python process and handles process metric
collection and reporting.

The tracer upholds the OpenTracing API and is responsible for reporting
span data to Instana.
"""

__author__ = 'Instana Inc.'
__copyright__ = 'Copyright 2017 Instana Inc.'
__credits__ = ['Pavlo Baron', 'Peter Giacomo Lombardo']
__license__ = 'MIT'
__version__ = get_distribution('instana').version
__maintainer__ = 'Peter Giacomo Lombardo'
__email__ = 'peter.lombardo@instana.com'


def load(module):
    """
    Method used to activate the Instana sensor via AUTOWRAPT_BOOTSTRAP
    environment variable.
    """
    if "INSTANA_DEV" in os.environ:
        print("==========================================================")
        print("Instana: Loading...")
        print("==========================================================")


# Optional application wide service name.
# Can be configured via environment variable or via code:
#
# export INSTANA_SERVICE_NAME=myservice
#   or
# instana.service_name = "myservice"
service_name = None

# User configurable EUM API key for instana.helpers.eum_snippet()
eum_api_key = ''

if "INSTANA_SERVICE_NAME" in os.environ:
    service_name = os.environ["INSTANA_SERVICE_NAME"]

if "INSTANA_DISABLE_AUTO_INSTR" not in os.environ:
    # Import & initialize instrumentation
    # noqa: ignore=W0611
    from .instrumentation import urllib3  # noqa
    from .instrumentation import sudsjurko  # noqa
    from .instrumentation import mysqlpython  # noqa
    from .instrumentation.django import middleware  # noqa
